package dev.pfaff.altargui.debug;

import it.unimi.dsi.fastutil.longs.Long2IntOpenHashMap;
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;

public final class EventTreeProfiler implements Profiler, AutoCloseable {
	private final Long2IntOpenHashMap threadDepths = new Long2IntOpenHashMap();
	private final Long2ObjectOpenHashMap<BufferedWriter> threadTrees = new Long2ObjectOpenHashMap<>();
	private final Path dir;

	public EventTreeProfiler(Path dir) {
		this.dir = dir;
		try {
			Files.createDirectories(dir);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void push(String name) {
	}

	@Override
	public void pop() {
	}

	@Override
	public void replace(String name) {
	}

	@Override
	public boolean isEventEnabled(Class<? extends Event> eventClass) {
		return true;
	}

	@Override
	public void pushEvent(Event event) {
		var thread = Thread.currentThread();
		var threadId = thread.threadId();
		int depth;
		BufferedWriter writer;
		synchronized (this) {
			depth = this.threadDepths.addTo(threadId, 1);
			//noinspection DataFlowIssue
			writer = this.threadTrees.putIfAbsent(threadId, null);
		}
		//noinspection ConstantValue
		if (writer == null) {
			try {
				writer = Files.newBufferedWriter(this.dir.resolve(thread.getName()));
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
			// no synchronization should be needed, since there is already an entry in the map for the key.
			this.threadTrees.put(threadId, writer);
		}
		try {
			writer.append('\n');
			for (int i = 0; i < depth; i++) {
				writer.append("  ");
			}
			writer.append("- ");
			writer.append(event.toString());
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void popEvent() {
		var threadId = Thread.currentThread().threadId();
		synchronized (this) {
			this.threadDepths.addTo(threadId, -1);
		}
	}

	@Override
	public void close() throws IOException {
		try {
			this.threadTrees.forEach((_, writer) -> {
				try {
					writer.flush();
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			});
		} catch (RuntimeException e) {
			if (e.getCause() instanceof IOException e1) {
				throw e1;
			}
			throw e;
		}
	}
}
